/*
btinit.c - Release 0.1.0  Btree initialize database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
        Boston, MA  02111-1307, USA.
*/

#include <stdlib.h>
#include <malloc.h>
#include <bthash.h>

#define BLKHDR        8
#define RCDHDR        4
#define MAXBLKSZ  65536
#define MINRPN        3
#define MAXRPN     1023
#define MAXRCDLEN 32000

int BTGetBlkInit(HANDLE hFile, int blklen,
   unsigned char *blk)
   {
   BOOL   bResult;
   DWORD  nBytesToRead;
   DWORD  nBytesRead;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   nBytesToRead = (DWORD) blklen;
   bResult = (BOOL) ReadFile(hFile,
      blk,
      nBytesToRead,
      &nBytesRead,
      NULL);
   if (bResult && nBytesRead == nBytesToRead)
      return(0);
   else if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTGetBlkInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTGetBlkInit: read error\n");
      return(SYS_ERR + msgerr);
      } /* read error */
   else
      {
      fprintf(stderr,"BTGetBlkInit: read error "
         "(rdlen %d)\n", nBytesRead);
      return(BTREE_ERR + 107);
      } /* read error */
   } /* BTGetBlkInit */

int BTPutBlkInit(HANDLE hFile, int blklen,
   unsigned char *blk)
   {
   BOOL bResult;
   DWORD nBytesToWrite;
   DWORD nBytesWritten;
   LPCVOID lpBuffer;
   DWORD   msgerr;
   LPVOID  lpMsgBuf;
   nBytesToWrite = (DWORD) blklen;
   lpBuffer = (LPCVOID) blk;
   bResult = WriteFile(hFile,
      lpBuffer,
      nBytesToWrite,
      &nBytesWritten,
      NULL);
   if (bResult && nBytesWritten == nBytesToWrite)
      return(0);
   else if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTPutBlkInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      return(SYS_ERR + msgerr);
      } /* write error */
   else
      {
      fprintf(stderr,"BTPutBlkInit: write error "
         "(wrtlen %d)", nBytesWritten);
      return(BTREE_ERR + 108);
      } /* write error */
   return(0);
   } /* BTPutBlkInit */

int BTClsInit(HANDLE hndl)
   {
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   bResult = (BOOL) CloseHandle(hndl);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTClsInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTClsInit: close error\n");
      return(SYS_ERR + msgerr);
      } /* if close error */
   return(0);
   } /* BTClsInit */

int BTBldHdrInit(HANDLE hndl, int blksz, int nodesz,
   int rcdlen, int keylen,
   int keyofst, int dletofst)
   {
   int rslt;
   int hdrsz;
   btreefmt *blk;
   unsigned char *p,*q;
   hdrsz = blksz + ROOT;
   blk = (btreefmt *) malloc(hdrsz+16);
   if (blk == NULL)
      {
      fprintf(stderr,"BTBldHdrInit: malloc error\n");
      return(BTREE_ERR + 110);
      } /* malloc err */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + hdrsz;
   while (p < q) *p++ = '\0';
   strcpy(blk->signature,BT_SIGNATURE);
   blk->blksz    = blksz;
   blk->nodesz   = nodesz;   /* records per node */
   blk->rcdlen   = rcdlen;
   blk->keylen   = keylen;
   blk->keyofst  = keyofst;
   blk->dletofst = dletofst;
   blk->blkhdr   = BLKHDR;
   blk->rcdhdr   = RCDHDR;
   blk->rootofst = ROOT;
   p = (unsigned char *) blk;
   rslt = BTPutBlkInit(hndl,hdrsz,p);
   free(blk);
   return(rslt);
   } /* BTBldHdrInit */

int BTTruncInitOld(char *dbname,HANDLE *hndl)
   {
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      TRUNCATE_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTTruncInitOld: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTTruncInitOld: %s: "
	 "open read/write\n",
         dbname);
      fprintf(stderr,"BTTruncInitOld: "
         "CreateFile error\n");
      return(SYS_ERR + msgerr);
      } /* if open error */
   return(0);
   } /* BTTruncInitOld */

int BTOpnInitOld(char *dbname,HANDLE *hndl)
   {
   DWORD  msgerr;
   DWORD  rslt;
   HANDLE oldhndl;
   unsigned char buf[32];
   LPVOID lpMsgBuf;
   *hndl = NULL;
   oldhndl = CreateFile(dbname,
      GENERIC_READ,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (oldhndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTOpnInitOld: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTOpnInitOld: %s: "
	 "open error\n",
         dbname);
      return(SYS_ERR + msgerr);
      } /* open err */
   rslt = BTGetBlkInit(oldhndl,16,buf);
   if (rslt)
      {
      BTClsInit(oldhndl);
      return(rslt);
      } /* BTGetBlkInit err */
   rslt = strncmp((char *)buf,BT_SIGNATURE,16); 
   if (rslt)
      {
      fprintf(stderr,"BTOpnInitOld: filename %s "
         "for non-btree file\n",
         dbname);
      BTClsInit(oldhndl);
      return(BTREE_ERR + 106);
      } /* not btree file */
   else
      {
      rslt = BTClsInit(oldhndl);
      if (rslt) return(rslt);
      } /* close after good signature */
   return(BTTruncInitOld(dbname,hndl));
   } /* BTOpnInitOld */

int BTOpnInitNew(char *dbname, HANDLE *hndl)
   {
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      CREATE_NEW,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = (DWORD) GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTOpnInitNew: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTOpnInitNew: %s: open read/write\n",
         dbname);
      if (msgerr == ERROR_FILE_EXISTS)
         {
         fprintf(stderr,"BTOpnInitNew: %s: file exists\n",
            dbname);
         } /* file exists */
      return(SYS_ERR + msgerr);
      }
   return(0);
   } /* BTOpnInitNew */

int BTOpnInitAll(char *dbname, HANDLE *hndl)
   {
   int rslt;
   DWORD  msgerr;
   HANDLE oldhndl;
   unsigned char buf[32];
   LPVOID lpMsgBuf;
   *hndl = NULL;
   oldhndl = CreateFile(dbname,
      GENERIC_READ,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (oldhndl != INVALID_HANDLE_VALUE)
      {
      rslt = BTGetBlkInit(oldhndl,16,buf);
      if (rslt) return(rslt);
      rslt = strncmp((char *)buf,BT_SIGNATURE,16); 
      if (rslt)
	 {
         fprintf(stderr,"BTOpnInitAll: filename %s "
	    "for non-btree file\n",
	    dbname);
         BTClsInit(oldhndl);
         return(BTREE_ERR + 115);
	 } /* not btree file */
      else
	 {
         rslt = BTClsInit(oldhndl);
	 if (rslt) return(rslt);
         return(BTTruncInitOld(dbname,hndl));
	 } /* close after good signature */
      } /* if good open rdonly */
   fprintf(stderr,"Creating %s\n", dbname);
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      CREATE_NEW,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"BTOpnInitAll: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"BTOpnInitAll: %s: "
         "open read/write\n",
         dbname);
      return(SYS_ERR + msgerr);
      }
   return(0);
   } /* BTOpnInitAll */

int BTOpnInit(char *dbname, int truncflag,
   int blksz, int nodesz,
   int rcdlen, int keylen,
   int keyofst, int dletofst)
   {
   int rslt;
   HANDLE hndl;
   if (truncflag == INITOLD)
      rslt = BTOpnInitOld(dbname,&hndl);
   else if (truncflag == INITNEW)
      rslt = BTOpnInitNew(dbname,&hndl);
   else if (truncflag == INITALL)
      rslt = BTOpnInitAll(dbname,&hndl);
   if (rslt) return(rslt);

   rslt = BTBldHdrInit(hndl,blksz,nodesz,
      rcdlen,keylen,keyofst,dletofst);
   if (rslt)
      {
      BTClsInit(hndl);
      return(rslt);
      } /* if bad bldhdr */
   return(BTClsInit(hndl));
   } /* BTOpnInit */

int BTValidateInit(char *dbname, int truncflag,
   int nodesz, int rcdlen, int keylen,
   int keyofst, int dletofst)
   {
   int blksz;
   blksz = ((rcdlen + RCDHDR) * nodesz) + BLKHDR;
   if (blksz > MAXBLKSZ)
      {
      fprintf(stderr,"btinit: block size %d "
	 "exceeds maximum %d\n",
	 blksz, MAXBLKSZ);
      fprintf(stderr,"btinit: Adjust "
	 "records-per-node "
	 "and record-length\n");
      return(BTREE_ERR + 111);
      } /* blksz */
   if (keylen > rcdlen)
      {
      fprintf(stderr,"btinit: key-length %d "
         "exceeds record-length %d\n",
	 keylen, rcdlen);
      return(BTREE_ERR + 112);
      }
   if (keyofst > rcdlen - keylen)
      {
      fprintf(stderr,"btinit: key-offset %d "
         "exceeds the record-length %d\n",
	 keyofst, rcdlen);
      fprintf(stderr,"    minus the key-length %d\n",
	 keylen);
      return(BTREE_ERR + 113);
      }
   if (dletofst >= keyofst
      && dletofst < keyofst + keylen)
      {
      fprintf(stderr,"btinit: delete-offset %d "
	 "occupies key area %d to %d\n",
	 dletofst, keyofst, keyofst + keylen - 1);
      return(BTREE_ERR + 117);
      }
   return(BTOpnInit(dbname,truncflag,
      blksz,nodesz,rcdlen,
      keylen,keyofst,dletofst));
   } /* BTValidateInit */

int btinit(char *dbname, int truncflag,
   int nodesz, int rcdlen, int keylen,
   int keyofst, int dletofst)
   {
   if (truncflag == INITNEW
      || truncflag == INITOLD
      || truncflag == INITALL);
   else
      {
      fprintf(stderr,"btinit: invalid "
	 "truncate flag %d\n",
	 truncflag);
      return(BTREE_ERR + 114);
      } /* bad truncflag */
   if (nodesz < MINRPN || nodesz > MAXRPN)
      {
      fprintf(stderr,"btinit: invalid "
         "records-per-node %d\n",
         nodesz);
      return(BTREE_ERR + 101);
      }
   if (!(nodesz & 1))
      {
      fprintf(stderr,"btinit: records-per-node "
         "%d must be odd\n",
	 nodesz);
      return(BTREE_ERR + 102);
      }
   if (rcdlen < 1 || rcdlen > MAXRCDLEN)
      {
      fprintf(stderr,"btinit: invalid "
         "record-length %d\n",
         rcdlen);
      return(BTREE_ERR + 103);
      }
   if (keylen < 1 || keylen > MAXRCDLEN)
      {
      fprintf(stderr,"btinit: invalid "
         "key-length %d\n",
         keylen);
      return(BTREE_ERR + 104);
      }
   if (keyofst < 0 || keyofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"btinit: invalid "
         "key-offset %d\n",
         keyofst);
      return(BTREE_ERR + 105);
      }
   if (dletofst < 0 || dletofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"btinit: invalid "
         "delete-offset %d\n",
         dletofst);
      return(BTREE_ERR + 116);
      }
   return(BTValidateInit(dbname,truncflag,
      nodesz,rcdlen,keylen,
      keyofst,dletofst));
   } /* btinit */
